#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

#include "ht2030.h"

#define DEVICE "/dev/mem"
#define A460_PC104_IO_8_BASE (0xb2000000)
#define A460_PC104_IO_8_SIZE (0x10000)

static uint8_t *ht2030_base;
#define ht2030_readb(addr) ({*(volatile uint8_t *)(ht2030_base + addr);})
#define ht2030_writeb(addr, data) ({ \
			*(volatile uint8_t *)(ht2030_base + addr) = data;})

static void print_bin(uintptr_t addr)
{
	uint8_t val;
	int i;

	val = ht2030_readb(addr);
	for (i = HT2030_CHANNELES_PER_PORT - 1; 0 <= i; i--)
		putchar(!!(val & (1<<i)) + '0');
}

static void print_input_data(void)
{
	puts("PB7..PB0 PA7..PA0");

	print_bin(HT2030_PB);
	putchar(' ');
	print_bin(HT2030_PA);
	putchar('\n');
}

static void print_output_data(void)
{
	puts("PC7..PC0");

	print_bin(HT2030_PC);
	putchar('\n');
}

static void output_data(uint8_t data)
{
	ht2030_writeb(HT2030_PC, data);
}

static int xstrtoul(const char *str, unsigned long *val)
{
	char *endptr;

	if (*str == '\0')
		return -1;

	errno = 0;
	*val = strtoul(str, &endptr, 0);
	if (errno || *endptr)
		return -1;

	return 0;
}

static int xmmap(void)
{
	int fd;

	fd = open(DEVICE, O_RDWR | O_SYNC);
	if (fd < 0) {
		perror("open");
		return -1;
	}

	ht2030_base = mmap(0x0, A460_PC104_IO_8_SIZE, PROT_READ | PROT_WRITE,
			   MAP_SHARED, fd, A460_PC104_IO_8_BASE);
	if (ht2030_base == MAP_FAILED) {
		perror("mmap");
		close(fd);
		return -1;
	}

	close(fd);

	return 0;
}

static void xmunmap(void)
{
	munmap(ht2030_base, A460_PC104_IO_8_SIZE);
}

int main(int argc, char *argv[])
{
	unsigned long data;
	int ret;

	ret = xmmap();
	if (ret)
		return EXIT_FAILURE;
	ret = atexit(xmunmap);
	if (ret)
		return EXIT_FAILURE;

	switch (argc) {
	case 1:
		print_input_data();
		break;
	case 2:
		ret = xstrtoul(argv[1], &data);
		if (ret) {
			fprintf(stderr, "invalid output data: %s\n", argv[1]);
			return EXIT_FAILURE;
		}
		if (data > HT2030_PC_MAX) {
			fprintf(stderr, "out of range(data=%ld)\n", data);
			return EXIT_FAILURE;
		}
		output_data((unsigned char)data);
		print_output_data();
		break;
	default:
		fprintf(stderr, "Usage: %s [OUTPUT DATA]\n", argv[0]);
		return EXIT_FAILURE;
	}

	return EXIT_SUCCESS;
}
